package edu.georgetown.gucs.fingerprinter;

import java.io.File;
import java.util.*;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.ObjectBuilder;
import org.slf4j.LoggerFactory;



/**
 * This is a class that allows threaded parallel creation of fingerprints.
 *
 * @author Clay Shields
 */


public class BulkFingerprinter {

    private final int maxThread;
    private final List<Fingerprint> fpList;
    private int finishedThreadCount;
    private final ArrayList<Fingerprint> fingerprint = new ArrayList<>();

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BulkFingerprinter.class);

    /**
     * A constructor for when you want to limit the number of threads that are run
     *
     * @param fingerprinterName the name of the fingerprinter-derived class to use for fingerprinting
     * @param files             A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary        The dictionary to be used for fingerprinting
     * @param maxThread         limit on number of threads to be used
     */

    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary, int maxThread) {

        this.maxThread = maxThread;
        this.finishedThreadCount = 0;
        fpList = new ArrayList<>();
    }


    /**
     * A constructor for when you want to limit the number of threads that are run
     *
     * @param maxThread limit on number of threads to be used
     */

    public BulkFingerprinter(int maxThread) {

        this.maxThread = maxThread;
        this.finishedThreadCount = 0;
        fpList = new ArrayList<>();
    }

    /**
     * A constructor that will set the number of threads equal to the number of cores.
     *
     * @param fingerprinterName the name of the fingerprinter-derived class to use for fingerprinting
     * @param files             A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary        The dictionary to be used for fingerprinting
     */


    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary) {

        this.maxThread = Runtime.getRuntime().availableProcessors();
        this.finishedThreadCount = 0;
        fpList = new ArrayList<>();
    }


    /**
     * Allow a thread to return its results. This is used instead of a synchronized
     * data structure for performance reasons - we want to have many threads read at
     * once for matching
     */

    synchronized void putResults(List<Fingerprint> incoming) {

        for (Fingerprint anIncoming : incoming) {
            fpList.add(anIncoming);
        }

        this.finishedThreadCount++;
    }



    protected Dictionary dict;
    protected Fingerprinter fingerprinter;
    protected String fingerprinterName;

    public BulkFingerprinter(Dictionary dict, String fingerprinterName) {
        maxThread = 4;
        fpList = null;
        this.dict = dict;
        fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);
        fingerprinter.setDictionary(dict);
        this.fingerprinterName = fingerprinterName;

    }

    public List<Fingerprint> asyncBulkFing(File[] file) {

        
      List<Fingerprint> fing = Arrays.stream(file).parallel().map(f -> fingerprinter.computeFingerprint(f.getAbsolutePath()).get(0)).collect(Collectors.toList());

        for(Fingerprint filename : fing ) {

            System.out.println( filename.getFileName()+ " " + filename.getBase64Fingerprint() + "\n");

        }

        return fing;

    }

}