package edu.georgetown.gucs.matcher;

import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.utility.Global;
import org.apache.xerces.impl.dv.util.Base64;

import java.util.List;

/**
 * This is the base class for all fingerprint matchers. A FingerprintMatcher takes two fingerprints and compares them to
 * determines if their documents match. Depending on the matcher, either a numerical score, indicating the degree of
 * similarity between the two fingerprints' documents, or a boolean value, indicating whether the documents are the
 * same, is returned.
 *
 * @author Clay Shields
 */
public class FingerprintMatcher {

    /**
     * the minimum score needed to be considered a match
     */
    protected int minimum_score;


    protected int score = 0;
    protected String comparedFile;
    protected String compareFile;
    protected Fingerprint fp1;
    protected Fingerprint fp2;

    /**
     * Constructor that sets the minimum score to use for matching two fingerprints to zero
     */
    public FingerprintMatcher() {
        this.minimum_score = Global.MIN_SCORE;
    }

    /**
     * Constructor that sets the minimum score to use for matching two fingerprints
     *
     * @param score the minimum score needed to be considered a match
     */
    public FingerprintMatcher(int score) {
        this.minimum_score = score;
    }

    /**
     * Provides the name of the matcher used to determine in the two fingerprints have matching documents
     *
     * @return the string value of the matcher used
     */
    public String getMatcherName() {
        return this.getClass().getName();
    }

    /**
     * Sets the minimum score for these two fingerprints to be considered a match. By default the minimum score is 0 and
     * will therefore consider all documents as matching.
     *
     * @param score the minimum score for these fingerprints to be considered a match
     */
    public void setMinimumScore(int score) {
        this.minimum_score = score;
    }

    /**
     * Provides the minimum score for these two fingerprints to be considered a match. By default the minimum score is 0
     * and will therefore consider all documents as matching.
     *
     * @return the minimum score for these fingerprints to be considered a match
     */
    public int getMinimumScore() {
        return this.minimum_score;
    }

    /**
     * Determines if the two fingerprints' documents match;  must be over-ridden in each specific matcher
     *
     * @param fp1 the byte array containing the first fingerprint
     * @param fp2 the byte array containing the second fingerprint
     * @return the boolean value indicating if the two fingerprints have matching documents
     */
    public boolean match(byte[] fp1, byte[] fp2) {
        return false;
    }

    /**
     * Determines if these two fingerprints' documents match.
     * <p>
     * Calls <code>match(byte[], byte[])</code>, which by default returns false and must be over-ridden in each specific
     * matcher.
     *
     * @param base64fp1 the Base64 encoded string containing the first fingerprint
     * @param base64fp2 the Base64 encoded string containing the second fingerprint
     * @return the boolean value indicating if the two fingerprints have matching documents; by default, returns false.
     */
    public boolean match(String base64fp1, String base64fp2) {
        return match(Base64.decode(base64fp1), Base64.decode(base64fp2));
    }

    /**
     * Determines if these two fingerprints' documents match.
     * <p>
     * Calls <code>match(byte[], byte[])</code>, which by default returns false and must be over-ridden in each specific
     * matcher.
     *
     * @param fingerprintList the Base64 encoded string containing the first fingerprint
     * @return the boolean value indicating if the two fingerprints have matching documents; by default, returns false.
     */
    public boolean match(List<Fingerprint> fingerprintList) {

        System.out.println("ERROR CALLING WRONG MATCH IN GAP");
        return fingerprintList.size() < 0;
    }

    /**
     * Determines a similarity score for these two fingerprints; must be over-ridden in each specific matcher
     *
     * @param fp1 the byte array containing the first fingerprint
     * @param fp2 the byte array containing the second fingerprint
     * @return the similarity score between these two fingerprints, ranges from 0-100
     */
    public int getScore(byte[] fp1, byte[] fp2) {
        return Global.MIN_SCORE;
    }

    /**
     * Determines a similarity score for these two fingerprints;
     * over-ridden in GoogleAllPairs
     */
    public void getScore() {
    }

    public int getScore(String fing1, String fing2) {
        return Global.MIN_SCORE;
    }

    /**
     * Determines a similarity score for these two fingerprints.
     * <p>
     * Calls <code>getScore(byte[], byte[])</code>, which by default returns 0 and must be over-ridden in each specific
     * matcher.
     *
     * @param base64fp1 the Base64 encoded string containing the first fingerprint
     * @param base64fp2 the Base64 encoded string containing the second fingerprint
     * @return the similarity score between these two fingerprints, ranges from 0-100
     */
    public int getScoreXML(String base64fp1, String base64fp2) {
        return getScore(Base64.decode(base64fp1), Base64.decode(base64fp2));
    }

    public synchronized FingerprintMatcher getScore(Fingerprint fp1, Fingerprint fp2) {
        this.fp1 = fp1;
        this.fp2 = fp2;

       // Global.debug(fp1.getBase64Fingerprint(), fp2.getBase64Fingerprint());
        return this;
    }


}
