package edu.georgetown.gucs.matcher;

import java.io.File;
import java.io.IOException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.ObjectBuilder;

/**
 * Compares two XML fingerprints and provides a score for their degree of similarity. For accurate results, the two XML
 * fingerprints need be created using the same {@linkplain Fingerprinter}
 *
 * @author Lindsay Neubauer
 */
public class ScoreFingerprints {

    private final String fingerprint1;
    private final String fingerprint2;
    private String fingerprinter;
    private final FingerprintMatcher matcher;

    /**
     * Constructor that creates a matcher object and two Base64 encoded fingerprint string objects
     *
     * @param matcher the string matcher name to use to determine the score of these fingerprints
     * @param fpFile1 the string filename of the first fingerprint
     * @param fpFile2 the string filename of the second fingerprint
     */
    public ScoreFingerprints(String matcher, String fpFile1, String fpFile2) {
        this.fingerprinter = null;
        this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcher);
        this.fingerprint1 = getFingerprint(fpFile1);
        this.fingerprint2 = getFingerprint(fpFile2);
    }

    /**
     * Extracts the Base64 encoded fingerprint string from the given XML fingerprint digest file
     *
     * @param filename the string filename of the XML fingerprint digest containing the Base64 encoded fingerprint
     * @return the Base64 encoded string fingerprint to use for this matcher comparison
     */
    private String getFingerprint(String filename) {
        try {
            File xml = new File(filename);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);

            Element digest = doc.getRootElement();
            setFingerprinter(digest.getAttributeValue("fingerprinter").trim());

            return digest.getChildTextTrim("base64Fingerprint");
        } catch (JDOMException | IOException e) {
            //ExceptionHandler.handleExceptionExit(e);
        }
        return null;
    }

    /**
     * Checks that the fingerprinters of these two fingerprints are the same. Comparing fingerprints that were created
     * using different fingerprinters leads to incorrect results.
     *
     * @param temp_fingerprinter the string name of the fingerprinter to compare to previous or future fingerprinters
     */
    private void setFingerprinter(String temp_fingerprinter) {
        if (this.fingerprinter == null) {
            this.fingerprinter = temp_fingerprinter;
        } else if (!this.fingerprinter.equals(temp_fingerprinter)) {
          /*  ExceptionHandler.handleExceptionThrow(new Exception("ScoreFingerprints: "
                    + "Fingerprints use different fingerprinters (" + this.fingerprinter + " and " + temp_fingerprinter
                    + "). This causes incorrect results " + "comparing the two fingerprints."));
        */
        }
    }

    /**
     * Depending on the matcher, prints either the boolean match or integer similarity of these two fingerprints
     */
    public void printMatch() {
        String matcherName = this.matcher.getClass().getSimpleName();

        if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {
            System.out.println(matcherName + " match = " + this.matcher.match(this.fingerprint1, this.fingerprint2));
        } else {
            System.out.println(matcherName + " score = " + this.matcher.getScoreXML(this.fingerprint1, this.fingerprint2));
        }
    }

    /**
     * Depending on the matcher, returns string representations of either the boolean match or integer similarity of these
     * two fingerprints
     *
     * @return the string showing the match
     */
    public String getMatch() {
        String match = this.matcher.getClass().getSimpleName();

        if (match.equals("HashFingerprintMatcher") || match.equals("ExactFingerprintMatcher")) {
            match += " match = " + this.matcher.match(this.fingerprint1, this.fingerprint2);
        } else {
            match += " score = " + this.matcher.getScoreXML(this.fingerprint1, this.fingerprint2);
        }
        return match;
    }


}
