package edu.georgetown.gucs.tokenizers;

import java.util.*;
import java.util.stream.Collectors;

import edu.georgetown.gucs.utility.Global;
import net.sf.classifier4J.summariser.SimpleSummariser;
import edu.georgetown.gucs.utility.Token;

/**
 * Manipulates files as they are read. Settings are specified in an XML configuration file for a ParameterSearch and can
 * include:
 * <ul>
 * <li>null: makes no changes</li>
 * <li>deleteToken: deletes a given percentage of the tokens</li>
 * <li>deleteChar: deletes a given percentage of single characters</li>
 * <li>deleteWhitespace: deletes a given percentage of white space</li>
 * <li>deleteSection: deletes a given percentage of contiguous tokens</li>
 * <li>deleteHead: deletes a given percentage of contiguous tokens from the beginning of the file</li>
 * <li>deleteTail: deletes a given percentage of contiguous tokens from the end of the file</li>
 * <li>splitToken: inserts white space randomly inside a given percentage of tokens</li>
 * <li>changeToChar: changes a given percentage of characters to a random ascii lower case letter or number</li>
 * <li>changeToken: changes a given percentage of tokens to a random token from the dictionary</li>
 * <li>changeToPunc: changes a given percentage of characters to punctuation</li>
 * <li>summaryReduce: uses Classifier4J to create a document summary of a given percentage of the document size</li>
 * </ul>
 *
 * @author Clay Shields
 */
public class FileManglerTokenizer extends Tokenizer {

    private static final long serialVersionUID = 5750194440419510989L;


    private Random random;
    private List<Token> tokensList;
    private boolean nullProcess;
    private String tokensAsString;

    Map<String, Float> map = new HashMap<String, Float>() {
        {
            put("deleteToken", null);
            put("deleteSection", null);
            put("deleteHead", null);
            put("deleteTail", null);
            put("deleteChar", null);
            put("deleteWhitespace", null);
            put("changeToChar", null);
            put("changeToPunc", null);
            put("changeToken", null);
            put("splitToken", null);
            put("summaryReduce", null);
        }
    };

    List<Token> tokenVector = Collections.synchronizedList(new Vector<>());

    /**
     * Ensures that the parameters provided to the mangler are correctly formatted
     * @param mangler the current mangler that is going to be used
     * @param percent the percent that is going to be mangled
     */
    //TODO make sure this works
    public void validateMangler(String mangler, float percent) {

        if (percent < 0 || percent > Global.MAX_SCORE) {

            //return error

        }
        if (!map.containsKey(mangler)) {
            //return error
        }
        if (map.get(mangler) > 0) {
            //error mangler cannot be greater than 0

        }

        map.put(mangler, percent);

    }

    /**
     * Constructor that initializes the random number generator and clears the mangler settings
     */
    public FileManglerTokenizer() {
        this.random = new Random();
        this.nullProcess = true;
        this.tokensAsString = null;
        clearManglerSettings();
    }

    /**
     * Constructor that sets the mangler settings
     *
     * @param manglers the string of mangler settings; expects non-null mangler strings in the form:
     *                 "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
     */
    public FileManglerTokenizer(String manglers) {
        this.random = new Random();
        this.tokensAsString = null;
        this.nullProcess = false;
        parseSettings(manglers);
    }

    /**
     * Turns off (sets to zero) all the mangler settings for this tokenizer
     */
    private void clearManglerSettings() {
        for (Map.Entry<String, Float> maping : map.entrySet()) {

            maping.setValue(0f);

        }
    }


    /**
     * Enables the mangler with the given settings in this tokenizer
     *
     * @param manglers  the string of mangler settings; expects non-null mangler strings in the form:
     *                  "manglerName-manglerPercent", where the manglerPercent is between 1 and 100 inclusive
     * @param tokenList the list of string tokens to put through this mangler
     */
    public void enableMangler(String manglers, List<Token> tokenList) {
        this.tokensList = tokenList;
        this.nullProcess = false;
        parseSettings(manglers);
    }

    /**
     * Disables the mangler in this tokenizer
     */
    public void disableMangler() {
        this.nullProcess = true;
        clearManglerSettings();
    }

    /**
     * Sets the random number generator to use with the manglers that are set in this tokenizer
     *
     * @param random the random number generator to use with this tokenizer
     */
    public void setRNG(Random random) {
        this.random = random;
    }

    /**
     * Sets each mangler setting present in the given list; throws an error if the same mangler is present more than once
     *
     * @param manglers the string of mangler settings; expects non-null mangler strings in the form:
     *                 "manglerName-manglerPercent", where the manglerPercent is between 1 and 100 inclusive
     */
    private void parseSettings(String manglers) {
        clearManglerSettings();
        if (manglers.equalsIgnoreCase("null")) {
            disableMangler();
        } else {
            String[] manglerArray = manglers.split("\\s");

            for (String manglerInput : manglerArray) {


                if (!manglerInput.contains("-")) {

                    System.err.println("The mangler [ " + manglerInput + "] is not properly formatted.");
                    System.err.println("Please use the format: manglerName-percentToMangle");
                    System.exit(1);

                }

                String[] manglerSplit = manglerInput.split("-");
                String mangler = manglerSplit[0];
                float percent = Float.valueOf(manglerSplit[1]);

                validateMangler(mangler, percent);

            } // end parseSettings
        }
    }

    /**
     * Alters or eliminates certain tokens based on the given mangler settings
     */
    @SuppressWarnings("StringBufferMayBeStringBuilder")
    @Override
    public List<Token> tokenize(List<Token> tokens) {

        synchronized (tokenVector) {

            tokenVector.addAll(tokens.stream()
                    .filter(token -> token.name().length() > 0)
                    .collect(Collectors.toList()));

        }

        // Apply mangler settings
        if  (this.nullProcess) {

            return tokenVector;

        }

        for (Map.Entry<String, Float> maping : map.entrySet()) {

            if (!(maping.getValue() > 0)) {
                continue;
            }
            if(tokenVector.size() <= 0 ) {
                break;
            }

            Vector<Token> newTokenVector = new Vector<>();
            switch (maping.getKey()) {

                case "summaryReduce":
                    summaryReduce(newTokenVector);
                    tokenVector.clear();
                    break;

                case "splitToken":
                    splitToken(newTokenVector);
                    break;
                case "ChangeToPunc":
                    changeToPunc(newTokenVector);
                    break;
                case "changeToken":
                    changeToken(newTokenVector);
                    break;
                case "changeToChar":
                    changeToChar(newTokenVector);
                    break;
                case "deleteWhiteSpace":
                    deleteWhiteSpace(newTokenVector);
                    break;
                case "deleteChar":
                    deleteChar(newTokenVector);
                    break;
                case "deleteToken":
                    deleteToken();
                    break;
                case "deleteSection":
                    deleteSection();
                    break;
                case "deleteHead":
                    deleteHead();
                    break;
                case "deleteTail":
                    deleteTail();
                    break;
                default:
                    //some error message here
            }

            if ( newTokenVector.size() > 0 ) {  tokenVector.addAll(newTokenVector); }

        }

            return tokenVector;

    }




    public void deleteToken(){

        Float deleteToken = map.get("deleteToken");

        // Deletes a given percentage of the tokens
        if (deleteToken > 0) {
            //((float) 90) / ((float) 100)
            int deletionSize = (int) ((deleteToken / (float) 100) * tokenVector.size());
            for (int i = 0; i < deletionSize; i++) {
                tokenVector.remove(this.random.nextInt(tokenVector.size()));
            }
        }

    }


    public void deleteSection(){

        Float deleteSection = map.get("deleteSection");

        int deletionSize = (int) ((deleteSection / 100) * tokenVector.size());
        int maxStart = tokenVector.size() - deletionSize;

        if (maxStart < 0) {
            String exception = "FileManglerTokenizer: DeleteSection is larger than token vector" + ": vector size= "
                    + tokenVector.size() + " setting in pct " + deleteSection + " num to delete " + deletionSize;

        }
        int randomStart = (maxStart == 0) ? 0 : this.random.nextInt(maxStart);
        tokenVector.subList(randomStart, randomStart + deletionSize).clear();

    }



    public void deleteHead(){

        Float deleteHead = map.get("deleteTail");

        int deletionSize = (int) ((deleteHead / 100) * tokenVector.size());
        tokenVector.subList(0, deletionSize).clear();

    }

    public void deleteTail(){

        Float deleteTail = map.get("deleteTail");

        int deletionSize = (int) ((deleteTail / 100) * tokenVector.size());
        tokenVector.subList(tokenVector.size() - deletionSize, tokenVector.size()).clear();


    }

    public void deleteChar(Vector<Token> newVectorToken){

        Float deleteChar = map.get("deleteChar");

        int deletionSize = (int) ((deleteChar / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {

            Token token = tokenVector.remove(random.nextInt(tokenVector.size()));
            StringBuilder stringBuffer = new StringBuilder(token.name());

            if (stringBuffer.length() > 0) {

                stringBuffer.deleteCharAt(random.nextInt(token.name().length()));
                newVectorToken.add(new Token(stringBuffer.toString(), token.start_byte(), token.end_byte()));

            }
        }

    }

    public void deleteWhiteSpace(Vector<Token> newTokenVector){

        Float deleteWhitespace = map.get("deleteWhiteSpace");

        int deletionSize = (int) ((deleteWhitespace / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
            int position = random.nextInt(tokenVector.size() - 1);
            // For each gap between tokens, there is a deleteWhitespace percent chance of it being deleted
            String newName = tokenVector.get(position).name() + tokenVector.get(position + 1).name();
            newTokenVector.set(position,
                    new Token(newName, tokenVector.get(position).start_byte(), tokenVector.get(position + 1).end_byte()));
            newTokenVector.remove(position + 1);
        }

    }

    public void changeToChar(Vector<Token> newTokenVector){

        Float changeChar = map.get("changeToChar");

        int deletionSize = (int) ((changeChar / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {

            Token token = tokenVector.remove(random.nextInt(tokenVector.size()));
            StringBuilder stringBuffer = new StringBuilder(token.name());
            stringBuffer.setCharAt(random.nextInt(stringBuffer.length()), (char) (random.nextInt(26) + 97));

            newTokenVector.add(new Token(stringBuffer.toString(), token.start_byte(), token.end_byte()));
        }


    }

    public void splitToken(Vector<Token> newTokenVector) {

        Float splitToken = map.get("splitToken");

        int deletionSize = (int) ((splitToken / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
            Token token = tokenVector.remove(this.random.nextInt(tokenVector.size()));
            if (token.name().length() > 2) {
                int splitPoint = this.random.nextInt(token.name().length());
                // Don't want to split the same token twice
                newTokenVector.add(new Token(token.name().substring(0, splitPoint), token.start_byte(), token.end_byte()));
                newTokenVector.add(new Token(token.name().substring(splitPoint), token.start_byte(), token.end_byte()));
            } else {
                newTokenVector.add(token);
            }
        }


    }

    public void changeToken(Vector<Token> newTokenVector){

        Float changeToken = map.get("changeToken");

        if (this.tokensList == null || this.tokensList.size() == 0) {
           return;
        }

        int deletionSize = (int) ((changeToken / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
            newTokenVector.set(this.random.nextInt(tokenVector.size()),
                    this.tokensList.get(this.random.nextInt(this.tokensList.size())));
        }

    }



    public void changeToPunc(Vector<Token> newTokenVector){


        Float changePunc = map.get("changeToPunc");

        // create a vector of punctuation symbols
        Vector<Character> punc = new Vector<>();

        changePuncHelper(punc, 33, 47);
        changePuncHelper(punc, 58, 64);
        changePuncHelper(punc, 91, 96);
        changePuncHelper(punc, 123, 126);

        int deletionSize = (int) ((changePunc / 100) * tokenVector.size());
        for (int i = 0; i < deletionSize; i++) {
            Token token = tokenVector.remove(random.nextInt(tokenVector.size()));
            StringBuilder stringBuffer = new StringBuilder(token.name());

            stringBuffer.setCharAt(random.nextInt(stringBuffer.length()), punc.get(random.nextInt(punc.size())));

            newTokenVector.add(new Token(stringBuffer.toString(), token.start_byte(), token.end_byte()));
        }

    }

    public void changePuncHelper(Vector<Character> punc,
                                 int start,
                                 int end) {

        for(char i = (char) start; i <= end; i++ ) {

            punc.add(i);

        }


    }

    /**
     * @param newtokenVector passed by value
     */
    public void summaryReduce(Vector<Token> newtokenVector) {

        Float sumReduce = map.get("summaryReduce");

        // Figure out how many sentences to keep
        int sentenceToKeep = (int) (tokensToSentences(newtokenVector) * ((100 - sumReduce) / 100));

        SimpleSummariser summariser = new SimpleSummariser();
        String summary = summariser.summarise(this.tokensAsString, sentenceToKeep);

        for (String result : summary.split("\\s")) {
            if (result.length() > 0) {
                // FIXME FileManglerTokenizer: Losing byte information by using this mangler
                newtokenVector.add(new Token(result.trim(), -1, -1));
            }
        }

    }

    /**
     * Processes tokens into sentences for use with the summariser; Does special processing when first line contains
     * "Message-ID" (looks like an email); Puts period at end of each header line to clearly mark it as the end of line
     *
     * @param tokenVector the vector containing all the tokens to convert to sentences
     * @return the <code>int</code> number of sentences
     */
    private int tokensToSentences(Vector<Token> tokenVector) {
        // This is a hack: Probably should have rewritten the File Tokenizer to do this
        boolean inHeaders = (tokenVector.get(0).name().contains("Message-ID:"));
        StringBuilder fullFile = new StringBuilder(tokenVector.get(0).name());
        int sentenceCount = 0;

        for (int i = 1; i < tokenVector.size(); i++) {
            fullFile.append(" ").append(tokenVector.get(i));

            if (inHeaders) {
                // peek ahead at the next token to see if it has a :
                if ((i < tokenVector.size() - 2) && (tokenVector.get(i + 1).name().contains(":"))) {
                    // if so, add a . to the current line
                    fullFile.append(".");
                    sentenceCount++;
                }
                // If at the X-FileName line should be the end of the headers
                if ((tokenVector.get(i).name().contains("X-FileName:"))) {
                    // A token that has a . in it is probably the file name
                    if (tokenVector.get(i).name().contains(".")) {
                        inHeaders = false;
                        fullFile.append(".");
                        sentenceCount++;
                    }
                }
            }

            // Check if we should count a new sentence
            // When we see a token that ends with a . ! or ?
            int length = tokenVector.get(i).name().length();

            // If it is a blank line, skip this check
            if (length != 0) {
                // char last = tokenVector.get(i).charAt(length - 1);
                if ((tokenVector.get(i).name().charAt(length - 1) == '.')
                        || (tokenVector.get(i).name().charAt(length - 1) == '!')
                        || (tokenVector.get(i).name().charAt(length - 1) == '?')) {
                    sentenceCount++;
                }
            }
        }// end for i

        this.tokensAsString = fullFile.toString();
        return sentenceCount;
    }
}
