package edu.georgetown.gucs.tokenizers;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Token;

/**
 * Partitions a set of tokens into multiple pieces
 *
 * @author Lindsay Neubauer
 */
public class Splitter extends Tokenizer {

    private static final long serialVersionUID = 6472644228611047470L;
    private List<Token> tokenList;

    /**
     * Constructor that takes a string indicating the type and percent for the splitter and a list of tokens to split
     *
     * @param splitter the string containing the splitter name and percent
     * @param list     the list of tokens to split
     */
    public Splitter(String splitter, List<Token> list) {
        //Global.monitor.writeToBuffer("Splitter", "constructor", "splitter=" + splitter + "  tokenList size=" + list.size());
        this.tokenList = Collections.synchronizedList(list);
        if (splitter == null || splitter.equalsIgnoreCase("null")) {
            this.tokenVectorMap.put("null", this.tokenList);
        } else {
            String[] split = splitter.split("\\s");
            String name = split[0];
            double value = Double.valueOf(split[1]);

            if (name.contains("proportions")) {
                proportion(value);
            } else if (name.contains("fixed")) {
                fixed((int) value);
            } else if (name.contains("w-shingle-unique")) {
                w_shingleUnique((int) value);
            } else if (name.contains("w-shingle")) {
                w_shingle((int) value);
            } else {
                 ExceptionHandler.handleExceptionThrow(new Exception(String.format("%s%n%s%n%s%n%s%n%s%n%s%n",
                         "Splitter: Unknown splitter (" + splitter + ") provided.",
                         "Please provide one of the following splitters in the XML configuration file:", " - proportions",
                         " - fixed", " - w-shingle", " - w-shingle-unique")));
            }
        }
    }

    /**
     * Splits the document into chunks based on the given proportion. Values between 0 and 1 split the document by a
     * percentage. Values >= 1 are truncated and split the document into that many parts.
     *
     * @param proportion the <code>double</code> proportion to split the document into; values between 0 and 1 split the document
     *                   by that percentage, values >= 1 are truncated and split the document into that many parts
     */
    private void proportion(double proportion) {
        int incrementSize = (proportion < 1) ? (int) (this.tokenList.size() * proportion + 1) : (int) (this.tokenList
                .size() / proportion + 1);
        if (incrementSize >= this.tokenList.size()) {
      /*      Global.monitor.printErrorMessage("Splitter", "Given split (proportion " + proportion
                    + ") exceeds the number of tokens in the document. Creating one chunk with size " + (this.tokenList.size()));
            Global.monitor.writeToBuffer("Splitter", "proportion", "Given split (proportion " + proportion
                    + ") exceeds the number of tokens in the document. Creating one chunk with size " + (this.tokenList.size()));
     */
        }
        for (int i = 0; i < this.tokenList.size(); i += incrementSize) {
            int end = (i + incrementSize <= this.tokenList.size()) ? i + incrementSize : this.tokenList.size();
            this.tokenVectorMap.put("proportion-" + proportion + "_" + (i / incrementSize + 1),
                    this.tokenList.subList(i, end));
        }
        //Global.monitor.writeToBuffer("Splitter", "proportion", "Completed");
    }

    /**
     * Splits the document into chunks of tokens of the given size. One chunk is created if the given number of tokens is
     * larger than the number of tokens in the document.
     *
     * @param chunkSize the <code>int</code> number of tokens to put in each portion of the document
     */
    private void fixed(int chunkSize) {
        int numTokens = chunkSize;
        if (numTokens >= this.tokenList.size()) {
          /*  Global.monitor.printErrorMessage("Splitter", "Given chunk size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one chunk with size " + (this.tokenList.size()));
            Global.monitor.writeToBuffer("Splitter", "fixed", "Given chunk size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one chunk with size " + (this.tokenList.size()));
            */
            numTokens = this.tokenList.size();
            //Global.monitor.writeToBuffer("Splitter", "fixed", "Updated numTokens=" + numTokens);
        }

        for (int i = 0; i < this.tokenList.size(); i += numTokens) {
            int end = (i + numTokens <= this.tokenList.size()) ? i + numTokens : this.tokenList.size();
            this.tokenVectorMap.put("fixed-" + numTokens + "_" + (i / numTokens + 1), this.tokenList.subList(i, end));
        }
    }

    /**
     * Splits the document into chunks using w-shingling. Each chunk is a contiguous subsequence of tokens in the document
     * of the given size. One chunk is created if the given number of tokens is larger than the number of tokens in the
     * document.
     *
     * @param nGrams the <code>int</code> number of tokens to put in each portion of the document
     */
    private void w_shingle(int nGrams) {
        if (nGrams >= this.tokenList.size()) {
      /*      Global.monitor.printErrorMessage("Splitter", "Given n-gram size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one shingle with n-gram = size = "
                    + (this.tokenList.size()));
            Global.monitor.writeToBuffer("Splitter", "w_shingle", "Given n-gram size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one shingle with n-gram = size = "
                    + (this.tokenList.size()));
            numTokens = this.tokenList.size();
            Global.monitor.writeToBuffer("Splitter", "w_shingle", "Updated nGrams=" + numTokens);
*/
        }

        int counter = 1;
        for (int i = 0; i <= this.tokenList.size() - nGrams; i++) {
            int end = (i + nGrams <= this.tokenList.size()) ? i + nGrams : this.tokenList.size();
            this.tokenVectorMap.put("w_shingle-" + nGrams + "_" + counter, this.tokenList.subList(i, end));
            counter++;
        }
    }

    /**
     * Splits the document into chunks using unique w-shingling. Each chunk is a contiguous and unique subsequence of
     * tokens in the document of the given size. One chunk is created if the given number of tokens is larger than the
     * number of tokens in the document.
     * <p>
     * The first occurrence of the subsequence of tokens is stored.
     * <p>
     * Unique w-shingling occurs before other tokenizers are applied so duplicate shingles are possible.
     *
     * @param nGrams the <code>int</code> number of tokens to put in each portion of the document
     */
    private void w_shingleUnique(int nGrams) {
        int numTokens = nGrams;
        if (numTokens >= this.tokenList.size()) {
  /*          Global.monitor.printErrorMessage("Splitter", "Given n-gram size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one shingle with n-gram = size = "
                    + (this.tokenList.size() - 1));
            Global.monitor.writeToBuffer("Splitter", "w_shingle-unique", "Given n-gram size (" + numTokens
                    + ") exceeds the number of tokens in the document. Creating one shingle with n-gram = size = "
                    + (this.tokenList.size() - 1));
     */
            numTokens = this.tokenList.size() - 1;
            //Global.monitor.writeToBuffer("Splitter", "w_shingleUnique", "Updated nGrams=" + numTokens);
        }

        Set<String> tokenSet = new LinkedHashSet<>();
        int counter = 1;
        for (int i = 0; i <= this.tokenList.size() - numTokens; i++) {
            int end = (i + numTokens <= this.tokenList.size()) ? i + numTokens : this.tokenList.size();
            List<Token> newTokenVector = this.tokenList.subList(i, end);
            if (tokenSet.add(newTokenVector.toString())) {
                this.tokenVectorMap.put("w_shingle-unique-" + numTokens + "_" + counter, newTokenVector);
                counter++;
            }
        }
        //Global.monitor.writeToBuffer("Splitter", "w_shingleUnique", "Completed");
    }

    /**
     * Tokenizes each splitter entry with the list of tokenizers; assumes no FileTokenizers are in the list
     *
     * @param tokenizers the list of tokenizers
     * @return the map containing splitters and their corresponding list of tokens
     */
    public Map<String, List<Token>> tokenizeEachEntry(List<Tokenizer> tokenizers) {
        for (Entry<String, List<Token>> splitterEntry : this.tokenVectorMap.entrySet()) {

            List<Token> tokenVector = splitterEntry.getValue();
            for (Tokenizer tokenizer : tokenizers) {

                tokenVector = tokenizer.tokenize(tokenVector);
            }
            this.tokenVectorMap.put(splitterEntry.getKey(), tokenVector);
        }
        return this.tokenVectorMap;
    }
}
