package edu.georgetown.gucs.tokenizers;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Token;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.*;


/**
 * @author Clay Shields
 * @author Lindsay Neubauer
 * @author Evan B.
 * @since  3/6/15.
 */
//awk -F '=' '{ print $2 }' enron_jprofile_v2 | cut -d "\"" -f2 | sort > token_jprofile
public class Tokenizer implements Serializable, Iterable<Map.Entry<String, List<Token>>> {

    private static final long serialVersionUID = 2475577088718479996L;

    /** The map containing the splitter name and its corresponding list of token elements */
      protected Map<String, List<Token>> tokenVectorMap = Collections.synchronizedMap(new TreeMap<>());

    /**
     * Alters or eliminates certain tokens. By default, this method adds all string tokens from the given iterator to the
     * list of tokens, and must be over-ridden in each specific tokenizer.
     *
     * @param tokenVector
     *          the vector of token elements to tokenize
     *
     * @return the vector of remaining tokens after tokenization
     */
    public List<Token> tokenize(List<Token> tokenVector) {
        this.tokenVectorMap.put("null", tokenVector);
        return tokenVector;
    }

    /**
     * Splits the document into tokens. By default, this method calls the ExceptionHandler to throw a
     * "Tokenizer does not support reading from a file" exception. It must be over-ridden in each tokenizer that supports
     * reading from a file.
     *
     * @param filename
     *          the filename of the document to split into tokens
     *
     * @return the vector of tokens present in the file
     */
    public List<Token> tokenize(String filename) {
        ExceptionHandler.handleExceptionThrow(new Exception("Tokenizer does not support reading from a file"));
        return null;
    }

    /**
     * Provides the list of each token in order of its appearance
     *
     * @return the map containing a splitter names and corresponding list of token elements
     */
    public Map<String, List<Token>> getTokenVectorMap() {
        return this.tokenVectorMap;
    }

    /**
     * Prints each token to the system output stream using the UTF-8 charset
     */
    public void printTokens() {
        try {
            OutputStreamWriter out = new OutputStreamWriter(System.out, Global.fileEncoding);
            BufferedWriter writer = new BufferedWriter(out);
            for (Map.Entry<String, List<Token>> entry : this.tokenVectorMap.entrySet()) {

                for (Token token : entry.getValue()) {
                    writer.write(String.format("%s%n", token));
                }
                writer.write(String.format("%n"));
            }
            out.flush();
            writer.flush();
        } catch (IOException e) {
            //log this error
        }
    }

    @Override
    public Iterator<Map.Entry<String, List<Token>>> iterator() {
        synchronized (this.tokenVectorMap) {
            return this.tokenVectorMap.entrySet().iterator();
        }
    }

    @Override
    public String toString() {
        return this.getClass().getName();
    }
}